#   Copyright 2015 Mirantis, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from unittest import mock
from unittest.mock import call

from novaclient import api_versions
from openstack import utils as sdk_utils
from osc_lib import exceptions

from openstackclient.compute.v2 import service
from openstackclient.tests.unit.compute.v2 import fakes as compute_fakes


class TestService(compute_fakes.TestComputev2):

    def setUp(self):
        super(TestService, self).setUp()

        self.app.client_manager.sdk_connection = mock.Mock()
        self.app.client_manager.sdk_connection.compute = mock.Mock()
        self.sdk_client = self.app.client_manager.sdk_connection.compute


class TestServiceDelete(TestService):

    services = compute_fakes.FakeService.create_services(count=2)

    def setUp(self):
        super(TestServiceDelete, self).setUp()

        self.sdk_client.delete_service.return_value = None

        # Get the command object to test
        self.cmd = service.DeleteService(self.app, None)

    def test_service_delete(self):
        arglist = [
            self.services[0].binary,
        ]
        verifylist = [
            ('service', [self.services[0].binary]),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.sdk_client.delete_service.assert_called_with(
            self.services[0].binary,
            ignore_missing=False
        )
        self.assertIsNone(result)

    def test_multi_services_delete(self):
        arglist = []
        for s in self.services:
            arglist.append(s.binary)
        verifylist = [
            ('service', arglist),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        calls = []
        for s in self.services:
            calls.append(call(s.binary, ignore_missing=False))
        self.sdk_client.delete_service.assert_has_calls(calls)
        self.assertIsNone(result)

    def test_multi_services_delete_with_exception(self):
        arglist = [
            self.services[0].binary,
            'unexist_service',
        ]
        verifylist = [
            ('service', arglist)
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        delete_mock_result = [None, exceptions.CommandError]
        self.sdk_client.delete_service = (
            mock.Mock(side_effect=delete_mock_result)
        )

        try:
            self.cmd.take_action(parsed_args)
            self.fail('CommandError should be raised.')
        except exceptions.CommandError as e:
            self.assertEqual(
                '1 of 2 compute services failed to delete.', str(e))

        self.sdk_client.delete_service.assert_any_call(
            self.services[0].binary,
            ignore_missing=False
        )
        self.sdk_client.delete_service.assert_any_call(
            'unexist_service',
            ignore_missing=False
        )


class TestServiceList(TestService):

    service = compute_fakes.FakeService.create_one_service()

    columns = (
        'ID',
        'Binary',
        'Host',
        'Zone',
        'Status',
        'State',
        'Updated At',
    )
    columns_long = columns + (
        'Disabled Reason',
    )

    data = [(
        service.id,
        service.binary,
        service.host,
        service.availability_zone,
        service.status,
        service.state,
        service.updated_at,
    )]
    data_long = [data[0] + (service.disabled_reason, )]

    def setUp(self):
        super(TestServiceList, self).setUp()

        self.sdk_client.services.return_value = [self.service]

        # Get the command object to test
        self.cmd = service.ListService(self.app, None)

    def test_service_list(self):
        arglist = [
            '--host', self.service.host,
            '--service', self.service.binary,
        ]
        verifylist = [
            ('host', self.service.host),
            ('service', self.service.binary),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.sdk_client.services.assert_called_with(
            host=self.service.host,
            binary=self.service.binary,
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))

    @mock.patch.object(sdk_utils, 'supports_microversion')
    def test_service_list_with_long_option(self, sm_mock):
        sm_mock.return_value = False

        arglist = [
            '--host', self.service.host,
            '--service', self.service.binary,
            '--long'
        ]
        verifylist = [
            ('host', self.service.host),
            ('service', self.service.binary),
            ('long', True)
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.sdk_client.services.assert_called_with(
            host=self.service.host,
            binary=self.service.binary,
        )

        self.assertEqual(self.columns_long, columns)
        self.assertEqual(self.data_long, list(data))

    @mock.patch.object(sdk_utils, 'supports_microversion')
    def test_service_list_with_long_option_2_11(self, sm_mock):
        sm_mock.return_value = True

        arglist = [
            '--host', self.service.host,
            '--service', self.service.binary,
            '--long'
        ]
        verifylist = [
            ('host', self.service.host),
            ('service', self.service.binary),
            ('long', True)
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.app.client_manager.compute.api_version = api_versions.APIVersion(
            '2.11')

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.sdk_client.services.assert_called_with(
            host=self.service.host,
            binary=self.service.binary,
        )

        # In 2.11 there is also a forced_down column.
        columns_long = self.columns_long + ('Forced Down',)
        data_long = [self.data_long[0] + (self.service.is_forced_down,)]

        self.assertEqual(columns_long, columns)
        self.assertEqual(data_long, list(data))


class TestServiceSet(TestService):

    def setUp(self):
        super(TestServiceSet, self).setUp()

        self.service = compute_fakes.FakeService.create_one_service()

        self.sdk_client.enable_service.return_value = self.service
        self.sdk_client.disable_service.return_value = self.service

        self.cmd = service.SetService(self.app, None)

    @mock.patch.object(sdk_utils, 'supports_microversion')
    def test_set_nothing(self, sm_mock):
        sm_mock.return_value = False
        arglist = [
            self.service.host,
            self.service.binary,
        ]
        verifylist = [
            ('host', self.service.host),
            ('service', self.service.binary),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.sdk_client.enable_service.assert_not_called()
        self.sdk_client.disable_service.assert_not_called()
        self.assertIsNone(result)

    @mock.patch.object(sdk_utils, 'supports_microversion')
    def test_service_set_enable(self, sm_mock):
        sm_mock.return_value = False
        arglist = [
            '--enable',
            self.service.host,
            self.service.binary,
        ]
        verifylist = [
            ('enable', True),
            ('host', self.service.host),
            ('service', self.service.binary),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.sdk_client.enable_service.assert_called_with(
            None,
            self.service.host,
            self.service.binary
        )
        self.assertIsNone(result)

    @mock.patch.object(sdk_utils, 'supports_microversion')
    def test_service_set_disable(self, sm_mock):
        sm_mock.return_value = False
        arglist = [
            '--disable',
            self.service.host,
            self.service.binary,
        ]
        verifylist = [
            ('disable', True),
            ('host', self.service.host),
            ('service', self.service.binary),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.sdk_client.disable_service.assert_called_with(
            None,
            self.service.host,
            self.service.binary,
            None
        )
        self.assertIsNone(result)

    @mock.patch.object(sdk_utils, 'supports_microversion')
    def test_service_set_disable_with_reason(self, sm_mock):
        sm_mock.return_value = False
        reason = 'earthquake'
        arglist = [
            '--disable',
            '--disable-reason', reason,
            self.service.host,
            self.service.binary,
        ]
        verifylist = [
            ('disable', True),
            ('disable_reason', reason),
            ('host', self.service.host),
            ('service', self.service.binary),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.sdk_client.disable_service.assert_called_with(
            None,
            self.service.host,
            self.service.binary,
            reason
        )
        self.assertIsNone(result)

    @mock.patch.object(sdk_utils, 'supports_microversion')
    def test_service_set_only_with_disable_reason(self, sm_mock):
        sm_mock.return_value = False
        reason = 'earthquake'
        arglist = [
            '--disable-reason', reason,
            self.service.host,
            self.service.binary,
        ]
        verifylist = [
            ('disable_reason', reason),
            ('host', self.service.host),
            ('service', self.service.binary),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        try:
            self.cmd.take_action(parsed_args)
            self.fail("CommandError should be raised.")
        except exceptions.CommandError as e:
            self.assertEqual("Cannot specify option --disable-reason without "
                             "--disable specified.", str(e))

    @mock.patch.object(sdk_utils, 'supports_microversion')
    def test_service_set_enable_with_disable_reason(self, sm_mock):
        sm_mock.return_value = False
        reason = 'earthquake'
        arglist = [
            '--enable',
            '--disable-reason', reason,
            self.service.host,
            self.service.binary,
        ]
        verifylist = [
            ('enable', True),
            ('disable_reason', reason),
            ('host', self.service.host),
            ('service', self.service.binary),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        try:
            self.cmd.take_action(parsed_args)
            self.fail("CommandError should be raised.")
        except exceptions.CommandError as e:
            self.assertEqual("Cannot specify option --disable-reason without "
                             "--disable specified.", str(e))

    @mock.patch.object(sdk_utils, 'supports_microversion')
    def test_service_set_state_up(self, sm_mock):
        sm_mock.side_effect = [False, True]
        arglist = [
            '--up',
            self.service.host,
            self.service.binary,
        ]
        verifylist = [
            ('up', True),
            ('host', self.service.host),
            ('service', self.service.binary),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        self.sdk_client.update_service_forced_down.assert_called_once_with(
            None,
            self.service.host,
            self.service.binary,
            False
        )
        self.assertNotCalled(self.sdk_client.enable_service)
        self.assertNotCalled(self.sdk_client.disable_service)
        self.assertIsNone(result)

    @mock.patch.object(sdk_utils, 'supports_microversion')
    def test_service_set_state_down(self, sm_mock):
        sm_mock.side_effect = [False, True]
        arglist = [
            '--down',
            self.service.host,
            self.service.binary,
        ]
        verifylist = [
            ('down', True),
            ('host', self.service.host),
            ('service', self.service.binary),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        self.sdk_client.update_service_forced_down.assert_called_once_with(
            None,
            self.service.host,
            self.service.binary,
            True
        )
        self.assertNotCalled(self.sdk_client.enable_service)
        self.assertNotCalled(self.sdk_client.disable_service)
        self.assertIsNone(result)

    @mock.patch.object(sdk_utils, 'supports_microversion')
    def test_service_set_enable_and_state_down(self, sm_mock):
        sm_mock.side_effect = [False, True]
        arglist = [
            '--enable',
            '--down',
            self.service.host,
            self.service.binary,
        ]
        verifylist = [
            ('enable', True),
            ('down', True),
            ('host', self.service.host),
            ('service', self.service.binary),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        self.sdk_client.enable_service.assert_called_once_with(
            None,
            self.service.host,
            self.service.binary
        )
        self.sdk_client.update_service_forced_down.assert_called_once_with(
            None,
            self.service.host,
            self.service.binary,
            True
        )
        self.assertIsNone(result)

    @mock.patch.object(sdk_utils, 'supports_microversion')
    def test_service_set_enable_and_state_down_with_exception(self, sm_mock):
        sm_mock.side_effect = [False, True]
        arglist = [
            '--enable',
            '--down',
            self.service.host,
            self.service.binary,
        ]
        verifylist = [
            ('enable', True),
            ('down', True),
            ('host', self.service.host),
            ('service', self.service.binary),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        with mock.patch.object(self.sdk_client, 'enable_service',
                               side_effect=Exception()):
            self.assertRaises(exceptions.CommandError,
                              self.cmd.take_action, parsed_args)
            self.sdk_client.update_service_forced_down.assert_called_once_with(
                None,
                self.service.host,
                self.service.binary,
                True
            )

    @mock.patch.object(sdk_utils, 'supports_microversion')
    def test_service_set_2_53_disable_down(self, sm_mock):
        # Tests disabling and forcing down a compute service with microversion
        # 2.53 which requires looking up the service by host and binary.
        sm_mock.return_value = True
        arglist = [
            '--disable',
            '--down',
            self.service.host,
            self.service.binary,
        ]
        verifylist = [
            ('disable', True),
            ('down', True),
            ('host', self.service.host),
            ('service', self.service.binary),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        service_id = '339478d0-0b95-4a94-be63-d5be05dfeb1c'
        self.sdk_client.services.return_value = [mock.Mock(id=service_id)]
        result = self.cmd.take_action(parsed_args)
        self.sdk_client.disable_service.assert_called_once_with(
            service_id,
            self.service.host,
            self.service.binary,
            None
        )
        self.sdk_client.update_service_forced_down.assert_called_once_with(
            service_id,
            self.service.host,
            self.service.binary,
            True)
        self.assertIsNone(result)

    @mock.patch.object(sdk_utils, 'supports_microversion')
    def test_service_set_2_53_disable_reason(self, sm_mock):
        # Tests disabling with reason a compute service with microversion
        # 2.53 which requires looking up the service by host and binary.
        sm_mock.return_value = True
        reason = 'earthquake'
        arglist = [
            '--disable',
            '--disable-reason', reason,
            self.service.host,
            self.service.binary,
        ]
        verifylist = [
            ('disable', True),
            ('disable_reason', reason),
            ('host', self.service.host),
            ('service', self.service.binary),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        service_id = '339478d0-0b95-4a94-be63-d5be05dfeb1c'
        self.sdk_client.services.return_value = [mock.Mock(id=service_id)]
        result = self.cmd.take_action(parsed_args)
        self.sdk_client.disable_service.assert_called_once_with(
            service_id,
            self.service.host,
            self.service.binary,
            reason
        )
        self.assertIsNone(result)

    @mock.patch.object(sdk_utils, 'supports_microversion')
    def test_service_set_2_53_enable_up(self, sm_mock):
        # Tests enabling and bringing up a compute service with microversion
        # 2.53 which requires looking up the service by host and binary.
        sm_mock.return_value = True
        arglist = [
            '--enable',
            '--up',
            self.service.host,
            self.service.binary,
        ]
        verifylist = [
            ('enable', True),
            ('up', True),
            ('host', self.service.host),
            ('service', self.service.binary),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        service_id = '339478d0-0b95-4a94-be63-d5be05dfeb1c'
        self.sdk_client.services.return_value = [mock.Mock(id=service_id)]
        result = self.cmd.take_action(parsed_args)
        self.sdk_client.enable_service.assert_called_once_with(
            service_id,
            self.service.host,
            self.service.binary
        )
        self.sdk_client.update_service_forced_down.assert_called_once_with(
            service_id,
            self.service.host,
            self.service.binary,
            False
        )
        self.assertIsNone(result)

    def test_service_set_find_service_by_host_and_binary_no_results(self):
        # Tests that no compute services are found by host and binary.
        self.sdk_client.services.return_value = []
        ex = self.assertRaises(exceptions.CommandError,
                               self.cmd._find_service_by_host_and_binary,
                               self.sdk_client, 'fake-host', 'nova-compute')
        self.assertIn('Compute service for host "fake-host" and binary '
                      '"nova-compute" not found.', str(ex))

    def test_service_set_find_service_by_host_and_binary_many_results(self):
        # Tests that more than one compute service is found by host and binary.
        self.sdk_client.services.return_value = [mock.Mock(), mock.Mock()]
        ex = self.assertRaises(exceptions.CommandError,
                               self.cmd._find_service_by_host_and_binary,
                               self.sdk_client, 'fake-host', 'nova-compute')
        self.assertIn('Multiple compute services found for host "fake-host" '
                      'and binary "nova-compute". Unable to proceed.',
                      str(ex))
